from .SingleThreadGenerator import SingleThreadGenerator as SingleThreadProjectionGenerator
from .OpenMPGenerator import OpenMPGenerator as OpenMPProjectionGenerator
from .CUDAGenerator import CUDAGenerator as CUDAProjectionGenerator